@@ -18,11 +18,6 @@ except ImportError:
     from debian_bundle import deb822, debfile, debian_support
 
 from config import TargetConfig
-# TODO(ajwong): Remove the from utils imports since style guide does not allow
-# importing of global functions in this manner.
-from utils import get_output
-from utils import spawn
-from utils import spawn_root
 import utils
 import aptutils
 import tree
@@ -58,8 +53,7 @@ def get_real_pkg(options, pkg):
     return tree.get_real_pkg(options, pkg)
 
 def get_src_name(options, pkg):
-    """Return the name of the source package that produces binary package
-    pkg."""
+    """Return the source package name that produces binary package pkg."""
 
     if options.apt_source or options.prefer_apt:
         src = aptutils.get_src_name(pkg)
@@ -97,13 +91,64 @@ def get_src_binaries(options, src):
         return tree.get_src_binaries(options, src)
 
 
+dpkg_architectures = None
+
+def dpkg_architecture_allowed(arch):
+    """Check if dpkg can install packages for the host architecture."""
+    global dpkg_architectures
+
+    if dpkg_architectures is not None:
+        return arch in dpkg_architectures
+
+    dpkg_architectures = set()
+    dpkg_architectures.add(
+        utils.get_output(['dpkg', '--print-architecture']).strip())
+    devnull = open('/dev/null', 'w')
+    dpkg_architectures.update(
+        utils.get_output(['dpkg', '--print-foreign-architectures'],
+                         mayfail=True, stderr=devnull).split())
+    devnull.close()
+
+    return arch in dpkg_architectures
+
+def apt_architecture_allowed(arch):
+    return (dpkg_architecture_allowed(arch) or
+            aptutils.apt_architecture_allowed(arch))
+
+
+def is_multiarch_foreign(options, pkg):
+    """Check if Multi-Arch: foreign is set on the binary package pkg.
+
+    Multi-Arch: foreign packages don't need to be built if they are just
+    build-deps and dpkg is configured to install packages of the appropriate
+    foreign architecture.
+    """
+
+    if not dpkg_architecture_allowed(options.architecture):
+        return False
+
+    real_pkg = get_real_pkg(options, pkg)
+    if real_pkg is None:
+        return False
+    record = get_pkg_record(options, real_pkg)
+    if record is None:
+       return False
+
+    if 'multi-arch' in record:
+        return record['multi-arch'] == 'foreign'
+    else:
+        return False
+
+
 def is_toolchain(pkg):
     """Is this package provided by the cross-toolchain?"""
-    """We really ought to check if they actually are installed or equived rather than assuming""" 
+
+    # We really ought to check if they actually are installed or equived
+    # rather than assuming
     return ('libc6' in pkg or 'lib32gcc' in pkg or 'lib64gcc' in pkg or
-            'libgcc' in pkg or 'libgcj' in pkg or 'libstdc++' in pkg or
-            'multilib' in pkg or
-            pkg == 'linux-libc-dev')
+            'libgcc' in pkg or 'libgcj' in pkg or 'lib32stdc++' in pkg or
+            'lib64stdc++' in pkg or 'libstdc++' in pkg or 'lib64c' in pkg or
+            'multilib' in pkg or pkg == 'linux-libc-dev')
 
 def is_crossable(pkg, conversion=False):
     """Can pkg sensibly be cross-built?"""
@@ -130,7 +175,8 @@ def is_crossable(pkg, conversion=False):
 
 def need_loop_break(parent, pkg):
     """Sometimes we need to break loops manually."""
-    if parent in ('python2.5', 'python2.6') and pkg == 'libbluetooth-dev':
+    if (parent in ('python2.5', 'python2.6', 'python2.7') and
+        pkg == 'libbluetooth-dev'):
         # Python build-depends on several things that are optional for
         # building. These are only used for module builds, and not for the
         # core Python binary and libraries that we need to cross-convert.
@@ -139,12 +185,33 @@ def need_loop_break(parent, pkg):
 
 
 explicit_requests = set()
+# Sets of source/binary packages that have been analysed
 all_srcs = {}
+all_pkgs = set()
 depends = {}
 needs_build = {}
+
+# set of graph edges in dependency tree
 dot_relationships = dict()
-needs_install = {}
-install_version = {}
+
+def print_relations(options, graph_pkg, parent, pkg, src, depth,
+                    parent_binary):
+    if options.generate_graph:
+        if options.generate_compact_graph:
+            rel = '"%s" -> "%s";' % ( parent, src)
+            dot_relationships[graph_pkg].append(rel)
+        else:
+            style = '"bin-%s" [label="%s",shape=box];' % ( pkg, pkg )
+            rel = '"%s" -> "bin-%s" -> "%s";' % ( parent, pkg, src)
+            dot_relationships[graph_pkg].append(style)
+            dot_relationships[graph_pkg].append(rel)
+    if options.debug:
+        if parent_binary is not None:
+            print "%s%s (%s) -> %s (%s)" % (' ' * depth,
+                                            parent, parent_binary,
+                                            src, pkg)
+        else:
+            print "%s%s -> %s (%s)" % (' ' * depth, parent, src, pkg)
 
 def filter_arch(deps, arch):
     """Filter out (build-)dependencies that do not apply to arch."""
@@ -180,22 +247,18 @@ def should_expand(options, parent, pkg):
         # We're going to build it anyway because it was requested on the
         # command line, so we might as well sequence it properly.
         return True
-     # if options.generate_graph:
-        # If generating non-compact graphs we want to report _real_ deps
-        # and not just ones that  xdeb approves of - however this just explodes dot
-        # disable for now
-        # return True;
     return (is_crossable(pkg) and not is_toolchain(pkg) and
+            not is_multiarch_foreign(options, pkg) and
             not need_loop_break(parent, pkg))
 
-def expand_depends(current_graph, options, pkg, depth, builddep_depth, 
+def expand_depends(options, graph_pkg, pkg, depth, builddep_depth,
                    parent=None, parent_binary=None):
     """Recursively expand (build-)dependencies of pkg."""
     # Get control data from apt or local dir
     # depth and builddep-depth are current level of recursion
-    # optionally output results in human readable (debug option), 
+    # optionally output results in human readable (debug option),
     #  or dot format (generate-graph option)
-    # 
+    #
     src = get_src_name(options, pkg)
     if not src:
         # Is it already a source package name?
@@ -220,89 +283,76 @@ def expand_depends(current_graph, options, pkg, depth, builddep_depth,
     src_record = get_src_record(options, src)
     if not src_record:
         if options.debug:
-            print "Did not find a source record for %s" % src 
+            print "Did not find a source record for %s" % src
         return
-    if options.debug:
-        print "expand_depends; parent=%s, package=%s, source=%s, depth=%s, builddepth=%s" % (parent, pkg, src, depth, builddep_depth)
-        print parent
 
     if parent is not None and parent != src:
-        if options.generate_graph:
-            if options.generate_compact_graph:
-                rel = '"%s" -> "%s";' % ( parent, src)
-                dot_relationships[current_graph].append(rel)
-            else:
-                style = '"bin-%s" [label="%s",shape=box];' % ( pkg, pkg )
-                rel = '"%s" -> "bin-%s" -> "%s";' % ( parent, pkg, src)
-                dot_relationships[current_graph].append(style)
-                dot_relationships[current_graph].append(rel)
-        if options.debug:
-            if parent_binary is not None:
-                print "%s%s (%s) -> %s (%s)" % (' ' * depth,
-                                                parent, parent_binary,
-                                                src, pkg)
-            else:
-                print "%s%s -> %s (%s)" % (' ' * depth, parent, src, pkg)
+        print_relations(options, graph_pkg, parent, pkg, src, depth,
+                        parent_binary)
         depends[parent].add(src)
-    if src in all_srcs:
-        return
-    all_srcs[src] = get_src_version(options, src)
-    depends[src] = set()
 
-    # In --only-explicit mode, we need to expand one level of
-    # build-dependencies and all their runtime dependencies in order that we
-    # can native-import them, but we don't want to go further down
-    # transitive build-dependencies.
-    if not options.only_explicit or not builddep_depth:
-        builddeps = src_record.relations['build-depends']
-        if options.architecture == build_arch:
-            builddeps.extend(src_record.relations['build-depends-indep'])
-        builddeps = filter_arch(builddeps, options.architecture)
-        for builddep in builddeps:
-            if [d for d in builddep if d['name'] == 'linux-gnu']:
-                # e.g. glib2.0 Build-Depends: libgamin-dev | libfam-dev | linux-gnu
-                # libgamin-dev Build-Depends: libglib2.0-dev, so we need to
-                # break this cycle
-                continue
-            bd_pkg = builddep[0]['name']
-            if should_expand(options, src, bd_pkg):
-                expand_depends(current_graph, options, bd_pkg, depth + 1, builddep_depth + 1,
-                               parent=src)
+    if src not in all_srcs:
+        # mark source as visited
+        all_srcs[src] = get_src_version(options, src)
+        depends[src] = set()
 
-    # Find the (install-)dependencies of the source package to ensure that 
-    # what is built can be installed. However we don't want to do this
-    # for a top-level package as that is not being installed (yet)
-    if depth:
-        for binary in get_src_binaries(options, src):
-            if not options.generate_graph:
-                if (not is_crossable(binary)) or is_toolchain(binary):
+        # In --only-explicit mode, we need to expand one level of
+        # build-dependencies and all their runtime dependencies in order
+        # that we can native-import them, but we don't want to go further
+        # down transitive build-dependencies.
+        if not options.only_explicit or not builddep_depth:
+            builddeps = src_record.relations['build-depends']
+            if options.architecture == build_arch:
+                builddeps.extend(src_record.relations['build-depends-indep'])
+            builddeps = filter_arch(builddeps, options.architecture)
+            for builddep in builddeps:
+                if [d for d in builddep if d['name'] == 'linux-gnu']:
+                    # e.g. glib2.0 Build-Depends: libgamin-dev | libfam-dev |
+                    #                             linux-gnu
+                    # libgamin-dev Build-Depends: libglib2.0-dev, so we need
+                    # to break this cycle
                     continue
-            parsed_binary = get_pkg_record(options, binary)
-            deps = parsed_binary.relations['pre-depends']
-            deps.extend(parsed_binary.relations['depends'])
-            filter_arch(deps, options.architecture)
-            for dep in deps:
-                for or_dep in dep:
-                    # Check should_expand twice, once for the virtual package
-                    # name and once for the real package name (which may be
-                    # different).
-                    if not should_expand(options, src, or_dep['name']):
-                        continue
-                    # TODO version handling?
-                    real_dep = get_real_pkg(options, or_dep['name'])
-                    if (real_dep is not None and
-                        get_src_name(options, real_dep) is not None):
-                        if should_expand(options, src, real_dep):
-                            expand_depends(current_graph, options, real_dep, depth + 1,
-                                           builddep_depth,
-                                           parent=src, parent_binary=binary)
-                        break
+                bd_pkg = builddep[0]['name']
+                if should_expand(options, src, bd_pkg):
+                    expand_depends(options, graph_pkg, bd_pkg, depth + 1,
+                                   builddep_depth + 1, parent=src)
+
+    # Find the (install-)dependencies of the source package to ensure that
+    # what is built can be installed. However we only want to do this
+    # for binary deps that are actually needed as build-deps
+    if (pkg not in all_pkgs and pkg in get_src_binaries(options, src) and
+        is_crossable(pkg) and not is_toolchain(pkg)):
+        all_pkgs.add(pkg)
+        parsed_binary = get_pkg_record(options, pkg)
+        deps = parsed_binary.relations['pre-depends']
+        deps.extend(parsed_binary.relations['depends'])
+        filter_arch(deps, options.architecture)
+        for dep in deps:
+            for or_dep in dep:
+                # Check should_expand twice, once for the virtual package
+                # name and once for the real package name (which may be
+                # different).
+                if not should_expand(options, src, or_dep['name']):
+                    continue
+                # TODO version handling?
+                real_dep = get_real_pkg(options, or_dep['name'])
+                if real_dep is None:
+                    continue
+                # only recurse further if src not already done
+                src_name = get_src_name(options, real_dep)
+                if src_name is None:
+                    continue
+                if should_expand(options, src, real_dep):
+                    expand_depends(options, graph_pkg, real_dep, depth + 1,
+                                   builddep_depth,
+                                   parent=src, parent_binary=pkg)
+                break
 
 def mark_needs_build(options, src, force=False):
     """Decide whether a package needs to be (re)built."""
     if src in needs_build:
         return
-    #moan if we end up getting a binary package name here
+    # moan if we end up getting a binary package name here
     assert src in all_srcs
     ver = all_srcs[src]
     newer = False
@@ -316,9 +366,6 @@ def mark_needs_build(options, src, force=False):
                     print "%s already built at version %s" % (src,
                                                               built_vers[-1])
                 newer = False
-                install_version[src] = built_vers[-1]
-                if check_if_src_needs_install(src, built_vers[-1], options):
-                    needs_install[src] = True
             else:
                 newer = True
         else:
@@ -360,38 +407,8 @@ def all_builds(options, src=None):
                 changes_file.close()
 
 
-apt_conf_written = False
-
-def update_apt_repository(options):
-    global apt_conf_written
-    apt_conf = os.path.join(options.destdir, 'apt.conf')
-    if not apt_conf_written:
-        apt_conf_file = open(apt_conf, 'w')
-        print >>apt_conf_file, '''
-Dir {
-	ArchiveDir ".";
-	CacheDir ".";
-};
-
-BinDirectory "." {
-	Packages "Packages";
-	BinCacheDB "pkgcache.apt";
-	FileList "filelist.apt";
-};'''
-        apt_conf_file.close()
-        apt_conf_written = True
-
-    filelist = os.path.join(options.destdir, 'filelist.apt')
-    filelist_file = open(filelist, 'w')
-    for name in sorted(os.listdir(options.destdir)):
-        if name.endswith('.deb'):
-            print >>filelist_file, './%s' % name
-    filelist_file.close()
-
-    spawn(['apt-ftparchive', 'generate', 'apt.conf'], cwd=options.destdir)
-
-
-build_arch = get_output(['dpkg-architecture', '-qDEB_BUILD_ARCH']).strip()
+build_arch = utils.get_output(['dpkg-architecture',
+                               '-qDEB_BUILD_ARCH']).strip()
 all_builddeps = set()
 
 # regexes from dak
@@ -406,6 +423,7 @@ def install_build_depends(options, srcs):
         if ('binutils-multiarch' in aptutils.cache and
             not aptutils.cache['binutils-multiarch'].is_installed):
             all_builddeps.add('binutils-multiarch')
+            available_builddeps.add('binutils-multiarch')
     for src in srcs:
         src_record = get_src_record(options, src)
         if src_record is None:
@@ -424,12 +442,23 @@ def install_build_depends(options, srcs):
             if (real_bd_pkg in aptutils.cache and
                 not aptutils.cache[real_bd_pkg].is_installed):
                 available_builddeps.add(real_bd_pkg)
+            if is_crossable(real_bd_pkg):
+                cross_bd_pkg = '%s-%s-cross' % (real_bd_pkg,
+                                                options.architecture)
+                if (cross_bd_pkg in aptutils.cache and
+                    not aptutils.cache[cross_bd_pkg].is_installed):
+                    available_builddeps.add(cross_bd_pkg)
 
     if available_builddeps:
-        command = ['apt-get', '-y', '--allow-unauthenticated',
-                   '--no-install-recommends', 'install']
+        command = ['apt-get', '-y']
+        command.extend(aptutils.apt_options(options))
+        command.extend(
+            ['-o', 'Dir::Etc::sourcelist=%s' %
+                   aptutils.sources_list_path(options),
+             '--no-install-recommends', 'install'])
         command.extend(sorted(available_builddeps))
-        spawn_root(command)
+        utils.spawn_root(command)
+        aptutils.reopen_cache()
 
 def cross_convert(options, debs, outdir='.'):
     crossable_debs = []
@@ -452,11 +481,11 @@ def cross_convert(options, debs, outdir='.'):
 
     crossed_debs = []
     if crossable_debs:
-        convert = ['dpkg-cross', '-a', options.architecture, '-b']
+        convert = ['dpkg-cross', '-a', options.architecture, '-A', '-M', '-b']
         for dep in exclude_deps:
             convert.extend(('-X', dep))
         convert.extend(crossable_debs)
-        spawn(convert, cwd=outdir)
+        utils.spawn(convert, cwd=outdir)
 
         # .debs use package name suffixed with -%arch-cross and arch: all;
         # \1 and \2 are package name and version from re_deb_filename
@@ -480,13 +509,20 @@ def native_import(options, src):
     print
 
     debs = []
+    previously_imported = False
     for binary in aptutils.get_src_binaries(src):
         if options.debug:
             print "Considering binary %s" % binary
+        command = ['apt-cache']
+        if apt_architecture_allowed(options.architecture):
+            command.append('-oAPT::Architecture=%s' % options.architecture)
+        command.append('show')
+        if apt_architecture_allowed(options.architecture):
+            command.append('%s:%s' % (binary, options.architecture))
+        else:
+            command.append(binary)
         try:
-            apt_opts = '-oAPT::Architecture=%s' % options.architecture
-            bin_cache = get_output(['apt-cache', apt_opts,
-                                    'show', binary]).splitlines()
+            bin_cache = utils.get_output(command).splitlines()
         except Exception:
             if options.debug:
                 print "skipping - %s" % binary
@@ -507,18 +543,28 @@ def native_import(options, src):
                 if deb_bits.group(3) == build_arch:
                     deb = re_deb_filename.sub(
                         r'\1_\2_%s.deb' % options.architecture, deb)
-                assert target_config.native_import_source, \
-                       "No native_import_source configured for arch %s" \
-                       % options.architecture
-                spawn(['wget', '-N',
-                       '%s/%s' % (target_config.native_import_source, deb)],
-                      cwd=options.builddirs[0])
+
+                if apt_architecture_allowed(options.architecture):
+                    command = ['apt-get',
+                               '-oAPT::Architecture=%s' % options.architecture,
+                               'download',
+                               '%s:%s' % (binary, options.architecture)]
+                else:
+                    assert target_config.native_import_source, \
+                           "No native_import_source configured for arch %s" \
+                           % options.architecture
+                    command = ['wget', '-N',
+                               '%s/%s' % (target_config.native_import_source,
+                                          deb)]
                 deb_base = deb.split('/')[-1]
+                if os.path.exists(os.path.join(options.destdir, deb_base)):
+                    previously_imported = True
+                utils.spawn(command, cwd=options.builddirs[0])
                 debs.append(deb_base)
             except StopIteration:
                 break
 
-    #fake up a changes file
+    # fake up a changes file
     changes = '%s_%s_%s.changes' % (src, ver_no_epoch, options.architecture)
     changes_file = open(os.path.join(options.builddirs[0], changes), 'w')
     print >>changes_file, 'Version: %s\nFake: yes' % ver
@@ -533,22 +579,15 @@ def native_import(options, src):
             os.rename(os.path.join(options.builddirs[0], f),
                       os.path.join(options.destdir, f))
 
-    update_apt_repository(options)
-
-    return crossed_debs
-
-def install_native_import(options, crossed_debs):
-    print "Installing native packages: %s" % crossed_debs
-    if crossed_debs:
-        install = ['dpkg', '-i']
-        install.extend(crossed_debs)
-        spawn_root(install, cwd=options.builddirs[0])
+    aptutils.update_apt_repository(options, force_rebuild=previously_imported)
 
 class BuildException(RuntimeError):
     pass
 
 def build(options, src, ver):
     """Build source package src at version ver."""
+    install_build_depends(options, [src])
+
     ver_no_epoch = re_no_epoch.sub('', ver)
     srcdir = tree.get_src_directory(options, src)
 
@@ -563,12 +602,12 @@ def build(options, src, ver):
                 refetch = True
 
     if use_apt and refetch:
-        spawn(['apt-get', '-d', 'source', '%s=%s' % (src, ver)],
-              cwd=options.builddirs[0])
+        utils.spawn(['apt-get', '-d', 'source', '%s=%s' % (src, ver)],
+                    cwd=options.builddirs[0])
         dsc = '%s_%s.dsc' % (src, ver_no_epoch)
         if srcdir:
             shutil.rmtree(srcdir, ignore_errors=True)
-        spawn(['dpkg-source', '-x', dsc, src], cwd=options.builddirs[0])
+        utils.spawn(['dpkg-source', '-x', dsc, src], cwd=options.builddirs[0])
         tree.scan_dir(os.path.join(options.builddirs[0], src))
         srcdir = tree.get_src_directory(options, src)
     else:
@@ -585,11 +624,14 @@ def build(options, src, ver):
         print "%s_%s not buildable for %s" % (src, ver, options.architecture)
         return
 
+    changes = '%s_%s_%s.changes' % (src, ver_no_epoch, options.architecture)
+    previously_built = os.path.exists(os.path.join(options.destdir, changes))
+
     print
     print "===== Building %s_%s =====" % (src, ver)
     print
 
-    spawn(['dpkg-checkbuilddeps'], cwd=srcdir)
+    utils.spawn(['dpkg-checkbuilddeps'], cwd=srcdir)
 
     buildpackage = ['debuild', '--no-lintian', '-eUSER']
     global target_config
@@ -607,9 +649,9 @@ def build(options, src, ver):
                                 os.environ['DEB_BUILD_OPTIONS'])
         else:
             buildpackage.append('-eDEB_BUILD_OPTIONS=nocheck')
-        deb_host_gnu_type = get_output(['dpkg-architecture',
-                                        '-a%s' % options.architecture,
-                                        '-qDEB_HOST_GNU_TYPE']).rstrip('\n')
+        deb_host_gnu_type = utils.get_output(
+            ['dpkg-architecture', '-a%s' % options.architecture,
+             '-qDEB_HOST_GNU_TYPE']).rstrip('\n')
         buildpackage.append('-eGTEST_INCLUDEDIR=/usr/%s/include' %
                             deb_host_gnu_type)
         buildpackage.append('-eGTEST_LIBDIR=/usr/%s/lib' % deb_host_gnu_type)
@@ -628,49 +670,22 @@ def build(options, src, ver):
     buildpackage.extend(['-b', '-uc', '-us'])
     if options.clean_after:
         buildpackage.append('-tc')
-    spawn(buildpackage, cwd=srcdir)
+    utils.spawn(buildpackage, cwd=srcdir)
 
     outdir = os.path.normpath(os.path.join(
         tree.get_src_directory(options, src), '..'))
-    install_built_pkg(src, ver_no_epoch, options, outdir)
-    
-def install_built_pkg(src, ver_no_epoch, options, outdir, force=True):
-    changes = '%s_%s_%s.changes' % (src, ver_no_epoch, options.architecture)
     build_log = '%s_%s_%s.build' % (src, ver_no_epoch, options.architecture)
-    debs = get_output(['dcmd', '--deb', changes], cwd=outdir).splitlines()
+    debs = utils.get_output(['dcmd', '--deb', changes],
+                            cwd=outdir).splitlines()
     print "Built packages:", ' '.join(debs)
 
-    # Install whatever we're going to need for future build-dependencies. We
-    # do this in both native and cross modes because it's fairly common to
-    # need things like data files from native packages even in cross mode,
-    # and because that allows us to build and install
-    # architecture-independent packages in cross mode.
-    builddep_debs = []
-    for deb in debs:
-        pkg = re_package.sub(r"\1", deb)
-        arch = re_package.sub(r"\3", deb)
-        # TODO: is_crossable is a hack, mostly to avoid problems with
-        # libfoo-dev depending on libfoo0 because we don't have a proper
-        # build-dependency installer yet.
-        if arch in (build_arch, 'all') and \
-            pkg not in target_config.blacklist and \
-            (pkg in all_builddeps or is_crossable(pkg) or
-             pkg in target_config.builddep_whitelist):
-            builddep_debs.append(deb)
-
     if options.architecture != build_arch:
         if options.lintian:
-            spawn(['lintian', '-C', 'xdeb', '-o', changes], cwd=outdir)
+            utils.spawn(['lintian', '-C', 'xdeb', '-o', changes], cwd=outdir)
         crossed_debs = cross_convert(options, debs, outdir)
-        builddep_debs.extend(crossed_debs)
-
-    if builddep_debs:
-        install = ['dpkg', '-i']
-        install.extend(builddep_debs)
-        spawn_root(install, cwd=outdir)
 
     if outdir != options.destdir:
-        files = get_output(['dcmd', changes], cwd=outdir).splitlines()
+        files = utils.get_output(['dcmd', changes], cwd=outdir).splitlines()
         for f in files:
             os.rename(os.path.join(outdir, f),
                       os.path.join(options.destdir, f))
@@ -682,33 +697,22 @@ def install_built_pkg(src, ver_no_epoch, options, outdir, force=True):
                 os.rename(os.path.join(outdir, deb),
                           os.path.join(options.destdir, deb))
 
-    update_apt_repository(options)
+    aptutils.update_apt_repository(options, force_rebuild=previously_built)
 
-def check_if_src_needs_install(src, ver, options):
-    ver_no_epoch = re_no_epoch.sub('', str(install_version[src]))
-    changes = '%s_%s_%s.changes' % (src, ver_no_epoch, options.architecture)
-    debs = get_output(['dcmd', '--deb', changes], cwd=options.destdir).splitlines()
-    install_reqd = False
-    for deb in debs:
-        pkg = re_package.sub(r"\1", deb)
-        if pkg in all_builddeps:
-            print "builddep: %s" % pkg
-            pkg_cross = "%s-%s-cross" % (pkg, options.architecture)
-            try:
-                if not aptutils.cache[pkg].is_installed:
-                    if options.debug:
-                        print ("% is not installed" % pkg)
-                    install_reqd = True
-                if not aptutils.cache[pkg_cross].is_installed:
-                    if options.debug:
-                        print ("% is not installed" % pkg)
-                    install_reqd = True
-            except:
-                print "failed to find pkg in cache:" + pkg
-                install_reqd = True
-    return install_reqd
-    
-def main():
+    if previously_built:
+        # We'll need to install all these again.
+        command = ['apt-get', '-y']
+        command.extend(aptutils.apt_options(options))
+        command.append('purge')
+        for deb in crossed_debs:
+            pkg = re_package.sub(r"\1", deb)
+            if pkg in aptutils.cache and aptutils.cache[pkg].is_installed:
+                command.append(pkg)
+        utils.spawn_root(command)
+        aptutils.reopen_cache()
+
+
+def parse_options():
     usage = '%prog [options] package ...'
     parser = optparse.OptionParser(usage=usage)
     parser.add_option('-C', '--config-files', dest='config_files',
@@ -780,10 +784,17 @@ def main():
                       dest='generate_graph', help='Generates a dot file that '
                                 'can be drawn by an app like GraphViz'
                                 'WARNING: WILL UNSET --debug')
-    parser.add_option('--generate-compact-graph', dest='generate_compact_graph',
+    parser.add_option('--generate-compact-graph',
+                      dest='generate_compact_graph',
                       action='store_true', default=False,
-                      help="Draw a simplified graph omitting binary package links.")
+                      help='draw a simplified graph omitting binary package '
+                           'links')
     options, args = parser.parse_args()
+    return parser, options, args
+
+
+def main():
+    parser, options, args = parse_options()
 
     config_paths = None
     if options.config_files:
@@ -795,11 +806,8 @@ def main():
     target_config.InitializeFromConfigs(config_paths)
     if options.generate_compact_graph:
         options.generate_graph = True
-    
+
     if options.generate_graph:
-        # Don't do real builds if generating graphs
-        # this doesn't really work - it prints stuff after graph
-        # options.sequence = True
         print '/*xdeb dot file starts here'
 
     if options.debug:
@@ -839,24 +847,27 @@ def main():
             print src, build_srcs[src]
         sys.exit(0)
 
+    aptutils.init(options)
+
     if options.convert:
         crossed_debs = cross_convert(options, args)
         if crossed_debs and options.convert_install:
             install = ['dpkg', '-i']
             install.extend(crossed_debs)
-            spawn_root(install)
+            utils.spawn_root(install)
         sys.exit(0)
 
     if options.all:
         args = tree.all_packages(options) + args
 
     explicit_requests.update(args)
+
     for pkg in args:
-        current_graph = None
+        graph_pkg = None
         if options.generate_graph:
             dot_relationships[pkg] = list()
-            current_graph = pkg
-        expand_depends(current_graph, options, pkg, 0, 0)
+            graph_pkg = pkg
+        expand_depends(options, graph_pkg, pkg, 0, 0)
 
     if options.generate_graph:
         print 'end of xdeb run output - graph starts here */'
@@ -871,8 +882,16 @@ def main():
 
     for pkg in args:
         if pkg not in all_srcs:
-            print "No source package found: %s" % pkg
-            sys.exit(1)
+            srcpkg = get_src_name(options, pkg)
+            if srcpkg in all_srcs:
+                print ("Using corresponding source %s for binary package %s" %
+                       (srcpkg, pkg))
+                args.remove(pkg)
+                if srcpkg not in args:
+                    args.append(srcpkg)
+            else:
+                print "No source or binary package found: %s" % pkg
+                sys.exit(1)
 
     if options.only_explicit:
         # In --only-explicit mode, we don't need to do a full topological
@@ -887,7 +906,6 @@ def main():
         except tsort.GraphCycleError, e:
             print "Dependency cycle:", e.graph
             sys.exit(1)
-    install_build_depends(options, build_sequence)
 
     for src in build_sequence:
         mark_needs_build(options, src)
@@ -917,20 +935,12 @@ def main():
     install_build_depends(options,
                           [src for src in build_sequence if src in real_build])
 
-    if options.debug:
-        print "Installing already built packages"
-    for src in needs_install:
-        install_built_pkg(src, install_version[src], options, 
-                                            options.destdir, False)
-    
     # In --only-explicit mode, native imports have no particular sequencing
     # requirements.
     if options.only_explicit:
-        native_crossed = []
         for src in build_sequence:
             if needs_build[src] and src not in real_build:
-                native_crossed.extend(native_import(options, src))
-        install_native_import(options, native_crossed)
+                native_import(options, src)
 
     for src in build_sequence:
         if options.debug:
@@ -939,7 +949,7 @@ def main():
             if src in real_build:
                 build(options, src, all_srcs[src])
             elif not options.only_explicit:
-                install_native_import(options, native_import(options, src))
+                native_import(options, src)
         else:
             if options.debug:
                 print "Skipping %s (already built)" % src
